@@ -15,11 +15,11 @@
 Test class for IPMI Module.
 """
 
-import mock
-import testtools
+from unittest import mock
 
 from pyghmi import exceptions as ipmi_exception
 from pyghmi.ipmi import command as ipmi_command
+import testtools
 
 from scciclient.irmc import ipmi
 
